@@ -107,7 +107,7 @@ PGAPI_Connect(
 
 	ci = &conn->connInfo;
 
-	make_string(szDSN, cbDSN, ci->dsn);
+	make_string(szDSN, cbDSN, ci->dsn, sizeof(ci->dsn));
 
 	/* get the values for the DSN from the registry */
 	memcpy(&ci->drivers, &globals, sizeof(globals));
@@ -120,8 +120,8 @@ PGAPI_Connect(
 	 * override values from DSN info with UID and authStr(pwd) This only
 	 * occurs if the values are actually there.
 	 */
-	make_string(szUID, cbUID, ci->username);
-	make_string(szAuthStr, cbAuthStr, ci->password);
+	make_string(szUID, cbUID, ci->username, sizeof(ci->username));
+	make_string(szAuthStr, cbAuthStr, ci->password, sizeof(ci->password));
 
 	/* fill in any defaults */
 	getDSNdefaults(ci);
@@ -277,6 +277,12 @@ CC_Constructor()
 		memset(rv->stmts, 0, sizeof(StatementClass *) * STMT_INCREMENT);
 
 		rv->num_stmts = STMT_INCREMENT;
+		rv->descs = (DescriptorClass **) malloc(sizeof(DescriptorClass *) * STMT_INCREMENT);
+		if (!rv->descs)
+			return NULL;
+		memset(rv->descs, 0, sizeof(DescriptorClass *) * STMT_INCREMENT);
+
+		rv->num_descs = STMT_INCREMENT;
 
 		rv->lobj_type = PG_TYPE_LO_UNDEFINED;
 
@@ -300,7 +306,8 @@ CC_Constructor()
 		rv->client_encoding = NULL;
 		rv->server_encoding = NULL;
 		rv->current_schema = NULL;
-
+		rv->num_discardp = 0;
+		rv->discardp = NULL;
 
 		/* Initialize statement options to defaults */
 		/* Statements under this conn will inherit these options */
@@ -332,6 +339,13 @@ CC_Destructor(ConnectionClass *self)
 		free(self->stmts);
 		self->stmts = NULL;
 	}
+
+	if (self->descs)
+	{
+		free(self->descs);
+		self->descs = NULL;
+	}
+
 	mylog("after free statement holders\n");
 
 	if (self->__error_message)
@@ -454,6 +468,7 @@ CC_cleanup(ConnectionClass *self)
 {
 	int			i;
 	StatementClass *stmt;
+	DescriptorClass *desc;
 
 	if (self->status == CONN_EXECUTING)
 		return FALSE;
@@ -490,6 +505,19 @@ CC_cleanup(ConnectionClass *self)
 		}
 	}
 
+	/* Free all the descs on this connection */
+	for (i = 0; i < self->num_descs; i++)
+	{
+		desc = self->descs[i];
+		if (desc)
+		{
+			DC_get_conn(desc) = NULL;	/* prevent any more dbase interactions */
+			DC_Destructor(desc);
+			free(desc);
+			self->descs[i] = NULL;
+		}
+	}
+
 	/* Check for translation dll */
 #ifdef WIN32
 	if (self->translation_handle)
@@ -520,8 +548,6 @@ CC_cleanup(ConnectionClass *self)
 	/* Free cached table info */
 	if (self->col_info)
 	{
-		int			i;
-
 		for (i = 0; i < self->ntables; i++)
 		{
 			if (self->col_info[i]->result)	/* Free the SQLColumns result structure */
@@ -535,6 +561,18 @@ CC_cleanup(ConnectionClass *self)
 		self->col_info = NULL;
 	}
 	self->ntables = 0;
+	if (self->num_discardp > 0 && self->discardp)
+	{
+		for (i = 0; i < self->num_discardp; i++)
+			free(self->discardp[i]);
+		self->num_discardp = 0;
+	}
+	if (self->discardp)
+	{
+		free(self->discardp);
+		self->discardp = NULL;
+	}
+
 	mylog("exit CC_Cleanup\n");
 	return TRUE;
 }
@@ -677,7 +715,11 @@ CC_connect(ConnectionClass *self, char password_req, char *salt_para)
 			return 0;
 		}
 
-		if (ci->server[0] == '\0' || ci->port[0] == '\0' || ci->database[0] == '\0')
+		if (ci->port[0] == '\0' ||
+#ifdef	WIN32
+			ci->server[0] == '\0' ||
+#endif /* WIN32 */
+			ci->database[0] == '\0')
 		{
 			CC_set_error(self, CONN_INIREAD_ERROR, "Missing server name, port, or database name in call to CC_connect.");
 			return 0;
@@ -916,7 +958,9 @@ another_version_retry:
 	mylog("sending an empty query...\n");
 
 	res = CC_send_query(self, " ", NULL, CLEAR_RESULT_ON_ABORT);
-	if (res == NULL || QR_get_status(res) != PGRES_EMPTY_QUERY)
+	if (res == NULL ||
+	    (QR_get_status(res) != PGRES_EMPTY_QUERY &&
+	     QR_command_nonfatal(res)))
 	{
 		mylog("got no result from the empty query.  (probably database does not exist)\n");
 		CC_set_error(self, CONNECTION_NO_SUCH_DATABASE, "The database does not exist on the server\nor user authentication failed.");
@@ -961,7 +1005,7 @@ another_version_retry:
 		CC_lookup_characterset(self);
 		if (CC_get_errornumber(self) != 0)
 			return 0;
-#ifdef UNICODE_SUPPORT
+		
 		if (self->unicode)
 		{
 			if (!self->client_encoding ||
@@ -985,18 +1029,13 @@ another_version_retry:
 				}
 			}
 		}
-#else
-		{
-		}
-#endif /* UNICODE_SUPPORT */
 	}
-#ifdef UNICODE_SUPPORT
 	else if (self->unicode)
 	{
 		CC_set_error(self, CONN_NOT_IMPLEMENTED_ERROR, "Unicode isn't supported before 6.4");
 		return 0;
 	}
-#endif /* UNICODE_SUPPORT */
+
 	ci->updatable_cursors = 0;
 #ifdef	DRIVER_CURSOR_IMPLEMENT
 	if (!ci->drivers.use_declarefetch &&
@@ -1052,9 +1091,11 @@ CC_remove_statement(ConnectionClass *self, StatementClass *stmt)
 {
 	int			i;
 
+	if (stmt->status == STMT_EXECUTING)
+		return FALSE;
 	for (i = 0; i < self->num_stmts; i++)
 	{
-		if (self->stmts[i] == stmt && stmt->status != STMT_EXECUTING)
+		if (self->stmts[i] == stmt)
 		{
 			self->stmts[i] = NULL;
 			return TRUE;
@@ -1147,6 +1188,32 @@ CC_get_error(ConnectionClass *self, int *number, char **message)
 	return rv;
 }
 
+static void CC_clear_cursors(ConnectionClass *self, BOOL allcursors)
+{
+	int	i;
+	StatementClass	*stmt;
+	QResultClass	*res;
+
+	for (i = 0; i < self->num_stmts; i++)
+	{
+		stmt = self->stmts[i];
+		if (stmt && (res = SC_get_Result(stmt)) &&
+			res->cursor && res->cursor[0])
+		{
+		/*
+		 * non-holdable cursors are automatically closed
+		 * at commit time.
+		 * all cursors are automatically closed
+		 * at rollback time.
+		 */
+			if (res->cursor)
+			{
+				free(res->cursor);
+				res->cursor = NULL;
+			}
+		}
+	}
+}
 
 void	CC_on_commit(ConnectionClass *conn)
 {
@@ -1160,9 +1227,15 @@ void	CC_on_commit(ConnectionClass *conn)
 		CC_set_no_manual_trans(conn);
 	}
 	conn->result_uncommitted = 0;
+	CC_clear_cursors(conn, TRUE);
+	CC_discard_marked_plans(conn);
 }
 void	CC_on_abort(ConnectionClass *conn, UDWORD opt)
 {
+	BOOL	set_no_trans = FALSE;
+
+	if (0 != (opt & CONN_DEAD))
+		opt |= NO_TRANS;
 	if (CC_is_in_trans(conn))
 	{
 #ifdef	DRIVER_CURSOR_IMPLEMENT
@@ -1173,8 +1246,10 @@ void	CC_on_abort(ConnectionClass *conn, UDWORD opt)
 		{
 			CC_set_no_trans(conn);
 			CC_set_no_manual_trans(conn);
+			set_no_trans = TRUE;
 		}
 	}
+	CC_clear_cursors(conn, TRUE);
 	if (0 != (opt & CONN_DEAD))
 	{
 		conn->status = CONN_DOWN;
@@ -1184,6 +1259,8 @@ void	CC_on_abort(ConnectionClass *conn, UDWORD opt)
 			conn->sock = NULL;
 		}
 	}
+	else if (set_no_trans)
+		CC_discard_marked_plans(conn);
 	conn->result_uncommitted = 0;
 }
 
@@ -1196,8 +1273,6 @@ void	CC_on_abort(ConnectionClass *conn, UDWORD opt)
  *	(i.e., C3326857) for SQL select statements.  This cursor is then used in future
  *	'declare cursor C3326857 for ...' and 'fetch 100 in C3326857' statements.
  */
-
-
 QResultClass *
 CC_send_query(ConnectionClass *self, char *query, QueryInfo *qi, UDWORD flag)
 {
@@ -1213,12 +1288,13 @@ CC_send_query(ConnectionClass *self, char *query, QueryInfo *qi, UDWORD flag)
 	int			maxlen,
 				empty_reqs;
 	BOOL		msg_truncated,
-				ReadyToReturn,
+				ReadyToReturn = FALSE,
 				query_completed = FALSE,
 				before_64 = PG_VERSION_LT(self, 6.4),
 				aborted = FALSE,
 				used_passed_result_object = FALSE;
 	UDWORD		abort_opt;
+	int		func_cs_count = 0;
 
 	/* ERROR_MSG_LENGTH is suffcient */
 	char msgbuffer[ERROR_MSG_LENGTH + 1];
@@ -1254,13 +1330,14 @@ CC_send_query(ConnectionClass *self, char *query, QueryInfo *qi, UDWORD flag)
 		return NULL;
 	}
 
-	ENTER_CONN_CS(self);
+#define	return DONT_CALL_RETURN_FROM_HERE???
+	ENTER_INNER_CONN_CS(self, func_cs_count);
 	SOCK_put_char(sock, 'Q');
 	if (SOCK_get_errcode(sock) != 0)
 	{
 		CC_set_error(self, CONNECTION_COULD_NOT_SEND, "Could not send Query to backend");
 		CC_on_abort(self, NO_TRANS | CONN_DEAD);
-		RETURN_AFTER_LEAVE_CS(self, NULL);
+		goto cleanup;
 	}
 
 	if (issue_begin)
@@ -1272,14 +1349,13 @@ CC_send_query(ConnectionClass *self, char *query, QueryInfo *qi, UDWORD flag)
 	{
 		CC_set_error(self, CONNECTION_COULD_NOT_SEND, "Could not send Query to backend");
 		CC_on_abort(self, NO_TRANS | CONN_DEAD);
-		RETURN_AFTER_LEAVE_CS(self, NULL);
+		goto cleanup;
 	}
 
 	mylog("send_query: done sending query\n");
 
-	ReadyToReturn = FALSE;
 	empty_reqs = 0;
-	for (wq = query; isspace((unsigned char) *wq); wq++)
+	for (wq = query; isspace((UCHAR) *wq); wq++)
 		;
 	if (*wq == '\0')
 		empty_reqs = 1;
@@ -1292,7 +1368,7 @@ CC_send_query(ConnectionClass *self, char *query, QueryInfo *qi, UDWORD flag)
 		if (!cmdres)
 		{
 			CC_set_error(self, CONNECTION_COULD_NOT_RECEIVE, "Could not create result info in send_query.");
-			RETURN_AFTER_LEAVE_CS(self, NULL);
+			goto cleanup;
 		}
 	}
 	res = cmdres;
@@ -1461,7 +1537,11 @@ CC_send_query(ConnectionClass *self, char *query, QueryInfo *qi, UDWORD flag)
 					abort_opt = NO_TRANS | CONN_DEAD;
 				}
 				else
+				{
 					CC_set_errornumber(self, CONNECTION_SERVER_REPORTED_WARNING);
+					if (CC_is_in_trans(self))
+						CC_set_in_error_trans(self);
+				}
 				CC_on_abort(self, abort_opt);
 				QR_set_status(res, PGRES_FATAL_ERROR);
 				QR_set_message(res, msgbuffer);
@@ -1568,6 +1648,8 @@ CC_send_query(ConnectionClass *self, char *query, QueryInfo *qi, UDWORD flag)
 		}
 	}
 
+cleanup:
+	CLEANUP_FUNC_CONN_CS(func_cs_count, self);
 	/*
 	 * Break before being ready to return.
 	 */
@@ -1616,7 +1698,8 @@ CC_send_query(ConnectionClass *self, char *query, QueryInfo *qi, UDWORD flag)
 			}
 		}
 	}
-	RETURN_AFTER_LEAVE_CS(self, retres);
+#undef	return
+	return retres;
 }
 
 
@@ -1701,6 +1784,8 @@ CC_send_function(ConnectionClass *self, int fnid, void *result_buf, int *actual_
 			case 'E':
 				SOCK_get_string(sock, msgbuffer, ERROR_MSG_LENGTH);
 				CC_set_errormsg(self, msgbuffer);
+				if (CC_is_in_trans(self))
+					CC_set_in_error_trans(self);
 				CC_on_abort(self, 0);
 
 				mylog("send_function(V): 'E' - %s\n", CC_get_errormsg(self));
@@ -1810,7 +1895,7 @@ CC_send_settings(ConnectionClass *self)
 	stmt->internal = TRUE;		/* ensure no BEGIN/COMMIT/ABORT stuff */
 
 	/* Set the Datestyle to the format the driver expects it to be in */
-	result = PGAPI_ExecDirect(hstmt, "set DateStyle to 'ISO'", SQL_NTS);
+	result = PGAPI_ExecDirect(hstmt, "set DateStyle to 'ISO'", SQL_NTS, 0);
 	if ((result != SQL_SUCCESS) && (result != SQL_SUCCESS_WITH_INFO))
 		status = FALSE;
 
@@ -1819,7 +1904,7 @@ CC_send_settings(ConnectionClass *self)
 	/* Disable genetic optimizer based on global flag */
 	if (ci->drivers.disable_optimizer)
 	{
-		result = PGAPI_ExecDirect(hstmt, "set geqo to 'OFF'", SQL_NTS);
+		result = PGAPI_ExecDirect(hstmt, "set geqo to 'OFF'", SQL_NTS, 0);
 		if ((result != SQL_SUCCESS) && (result != SQL_SUCCESS_WITH_INFO))
 			status = FALSE;
 
@@ -1830,7 +1915,7 @@ CC_send_settings(ConnectionClass *self)
 	/* KSQO (not applicable to 7.1+ - DJP 21/06/2002) */
 	if (ci->drivers.ksqo && PG_VERSION_LT(self, 7.1))
 	{
-		result = PGAPI_ExecDirect(hstmt, "set ksqo to 'ON'", SQL_NTS);
+		result = PGAPI_ExecDirect(hstmt, "set ksqo to 'ON'", SQL_NTS, 0);
 		if ((result != SQL_SUCCESS) && (result != SQL_SUCCESS_WITH_INFO))
 			status = FALSE;
 
@@ -1841,7 +1926,7 @@ CC_send_settings(ConnectionClass *self)
 	/* extra_float_digits (applicable since 7.4) */
 	if (PG_VERSION_GT(self, 7.3))
 	{
-		result = PGAPI_ExecDirect(hstmt, "set extra_float_digits to 2", SQL_NTS);
+		result = PGAPI_ExecDirect(hstmt, "set extra_float_digits to 2", SQL_NTS, 0);
 		if ((result != SQL_SUCCESS) && (result != SQL_SUCCESS_WITH_INFO))
 			status = FALSE;
 
@@ -1860,7 +1945,7 @@ CC_send_settings(ConnectionClass *self)
 #endif /* HAVE_STRTOK_R */
 		while (ptr)
 		{
-			result = PGAPI_ExecDirect(hstmt, ptr, SQL_NTS);
+			result = PGAPI_ExecDirect(hstmt, ptr, SQL_NTS, 0);
 			if ((result != SQL_SUCCESS) && (result != SQL_SUCCESS_WITH_INFO))
 				status = FALSE;
 
@@ -1887,7 +1972,7 @@ CC_send_settings(ConnectionClass *self)
 #endif /* HAVE_STRTOK_R */
 		while (ptr)
 		{
-			result = PGAPI_ExecDirect(hstmt, ptr, SQL_NTS);
+			result = PGAPI_ExecDirect(hstmt, ptr, SQL_NTS, 0);
 			if ((result != SQL_SUCCESS) && (result != SQL_SUCCESS_WITH_INFO))
 				status = FALSE;
 
@@ -1934,7 +2019,7 @@ CC_lookup_lo(ConnectionClass *self)
 		return;
 	stmt = (StatementClass *) hstmt;
 
-	result = PGAPI_ExecDirect(hstmt, "select oid from pg_type where typname='" PG_TYPE_LO_NAME "'", SQL_NTS);
+	result = PGAPI_ExecDirect(hstmt, "select oid from pg_type where typname='" PG_TYPE_LO_NAME "'", SQL_NTS, 0);
 	if ((result != SQL_SUCCESS) && (result != SQL_SUCCESS_WITH_INFO))
 	{
 		PGAPI_FreeStmt(hstmt, SQL_DROP);
@@ -2020,7 +2105,7 @@ CC_lookup_pg_version(ConnectionClass *self)
 	stmt = (StatementClass *) hstmt;
 
 	/* get the server's version if possible	 */
-	result = PGAPI_ExecDirect(hstmt, "select version()", SQL_NTS);
+	result = PGAPI_ExecDirect(hstmt, "select version()", SQL_NTS, 0);
 	if ((result != SQL_SUCCESS) && (result != SQL_SUCCESS_WITH_INFO))
 	{
 		PGAPI_FreeStmt(hstmt, SQL_DROP);
@@ -2116,7 +2201,7 @@ CC_get_max_query_len(const ConnectionClass *conn)
 }
 
 /*
- *	This deosn't really return the CURRENT SCHEMA
+ *	This doesn't really return the CURRENT SCHEMA
  *	but there's no alternative.
  */
 const char *
@@ -2139,38 +2224,30 @@ CC_get_current_schema(ConnectionClass *conn)
 int
 CC_send_cancel_request(const ConnectionClass *conn)
 {
-#ifdef WIN32
-	int			save_errno = (WSAGetLastError());
-#else
-	int			save_errno = errno;
-#endif
+	int			save_errno = SOCK_ERRNO;
 	int			tmpsock = -1;
 	struct
 	{
 		uint32		packetlen;
 		CancelRequestPacket cp;
 	}			crp;
+	BOOL	ret = TRUE;
 
 	/* Check we have an open connection */
-	if (!conn)
+	if (!conn || !conn->sock)
 		return FALSE;
 
-	if (conn->sock == NULL )
-	{
-		return FALSE;
-	}
-
 	/*
 	 * We need to open a temporary connection to the postmaster. Use the
 	 * information saved by connectDB to do this with only kernel calls.
 	*/
-	if ((tmpsock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
+	if ((tmpsock = socket(conn->sock->sadr->sa_family, SOCK_STREAM, 0)) < 0)
 	{
 		return FALSE;
 	}
-	if (connect(tmpsock, (struct sockaddr *)&(conn->sock->sadr),
-				sizeof(conn->sock->sadr)) < 0)
+	if (connect(tmpsock, conn->sock->sadr, conn->sock->sadr_len) < 0)
 	{
+		closesocket(tmpsock);
 		return FALSE;
 	}
 
@@ -2184,15 +2261,50 @@ CC_send_cancel_request(const ConnectionClass *conn)
 
 	if (send(tmpsock, (char *) &crp, sizeof(crp), 0) != (int) sizeof(crp))
 	{
-		return FALSE;
+		save_errno = SOCK_ERRNO;
+		ret = FALSE;
 	}
 
 	/* Sent it, done */
 	closesocket(tmpsock);
-#ifdef WIN32
-	WSASetLastError(save_errno);
-#else
-	errno = save_errno;
-#endif
-	return TRUE;
+	SOCK_ERRNO_SET(save_errno);
+
+	return ret;
+}
+
+int	CC_mark_a_plan_to_discard(ConnectionClass *conn, const char *plan)
+{
+	int	cnt = conn->num_discardp + 1;
+	char	*pname;
+
+	CC_REALLOC_return_with_error(conn->discardp, char *,
+		(cnt * sizeof(char *)), conn, "Couldn't alloc discardp.", -1) 
+	CC_MALLOC_return_with_error(pname, char, (strlen(plan) + 1),
+		conn, "Couldn't alloc discardp mem.", -1)
+	strcpy(pname, plan);
+	conn->discardp[conn->num_discardp++] = pname; 
+	return 1;
+}
+int	CC_discard_marked_plans(ConnectionClass *conn)
+{
+	int	i, cnt;
+	QResultClass *res;
+	char	cmd[32];
+
+	if ((cnt = conn->num_discardp) <= 0)
+		return 0;
+	for (i = cnt - 1; i >= 0; i--)
+	{
+		sprintf(cmd, "DEALLOCATE \"%s\"", conn->discardp[i]);
+		res = CC_send_query(conn, cmd, NULL, CLEAR_RESULT_ON_ABORT);
+		if (res)
+		{
+			QR_Destructor(res);
+			free(conn->discardp[i]);
+			conn->num_discardp--;
+		}
+		else
+			return -1;
+	}
+	return 1;
 }
