@@ -127,7 +127,7 @@ makeConnectString(char *connect_string, const ConnInfo *ci, UWORD len)
 	/* Abbrebiation is needed ? */
 	if (abbrev || strlen(connect_string) >= len)
 	{
-		unsigned long flag = 0;
+		UInt4 flag = 0;
 		if (ci->disallow_premature)
 			flag |= BIT_DISALLOWPREMATURE;
 		if (ci->allow_keyset)
@@ -140,6 +140,8 @@ makeConnectString(char *connect_string, const ConnInfo *ci, UWORD len)
 			flag |= BIT_PROTOCOL_64;
 		else if (strncmp(ci->protocol, PG63, strlen(PG63)) == 0)
 			flag |= BIT_PROTOCOL_63;
+		else if (strncmp(ci->protocol, "7.4", 3) == 0)
+			flag |= (BIT_PROTOCOL_63 | BIT_PROTOCOL_64);
 		switch (ci->drivers.unknown_sizes)
 		{
 			case UNKNOWNS_AS_DONTKNOW:
@@ -207,7 +209,7 @@ static void
 unfoldCXAttribute(ConnInfo *ci, const char *value)
 {
 	int		count;
-	unsigned long	flag;
+	UInt4		flag;
 
 	if (strlen(value) < 2)
 	{
