@@ -73,8 +73,8 @@ extern Int2 sqlTypes[];
 
 Int4		sqltype_to_pgtype(StatementClass *stmt, Int2 fSqlType);
 
-Int2		pgtype_to_concise_type(StatementClass *stmt, Int4 type);
-Int2		pgtype_to_sqldesctype(StatementClass *stmt, Int4 type);
+Int2		pgtype_to_concise_type(StatementClass *stmt, Int4 type, int col);
+Int2		pgtype_to_sqldesctype(StatementClass *stmt, Int4 type, int col);
 Int2		pgtype_to_datetime_sub(StatementClass *stmt, Int4 type);
 Int2		pgtype_to_ctype(StatementClass *stmt, Int4 type);
 const char	*pgtype_to_name(StatementClass *stmt, Int4 type);
@@ -88,6 +88,8 @@ Int4		pgtype_desclength(StatementClass *stmt, Int4 type, int col, int handle_unk
 Int4		pgtype_transfer_octet_length(StatementClass *stmt, Int4 type, int col, int handle_unknown_size_as);
 
 Int2		pgtype_decimal_digits(StatementClass *stmt, Int4 type, int col); /* corresponds to "scale" in ODBC 2.x */
+Int2		pgtype_min_decimal_digits(StatementClass *stmt, Int4 type); /* corresponds to "min_scale" in ODBC 2.x */
+Int2		pgtype_max_decimal_digits(StatementClass *stmt, Int4 type); /* corresponds to "max_scale" in ODBC 2.x */
 Int2		pgtype_scale(StatementClass *stmt, Int4 type, int col); /* ODBC 3.x " */
 Int2		pgtype_radix(StatementClass *stmt, Int4 type);
 Int2		pgtype_nullable(StatementClass *stmt, Int4 type);
