@@ -84,6 +84,8 @@ PGAPI_AllocStmt(HDBC hdbc,
 	CSTR func = "PGAPI_AllocStmt";
 	ConnectionClass *conn = (ConnectionClass *) hdbc;
 	StatementClass *stmt;
+	ARDFields	*ardopts;
+	BindInfoClass	*bookmark;
 
 	mylog("%s: entering...\n", func);
 
@@ -118,10 +120,9 @@ PGAPI_AllocStmt(HDBC hdbc,
 
 	/* Copy default statement options based from Connection options */
 	stmt->options = stmt->options_orig = conn->stmtOptions;
-	stmt->ardopts = conn->ardOptions;
-	stmt->ardopts.bookmark = (BindInfoClass *) malloc(sizeof(BindInfoClass));
-	stmt->ardopts.bookmark->buffer = NULL;
-	stmt->ardopts.bookmark->used = NULL;
+	stmt->ardi.ardopts = conn->ardOptions;
+	ardopts = SC_get_ARDF(stmt);
+	bookmark = ARD_AllocBookmark(ardopts);
 
 	stmt->stmt_size_limit = CC_get_max_query_len(conn);
 	/* Save the handle for later */
@@ -218,34 +219,7 @@ InitializeStatementOptions(StatementOptions *opt)
 	opt->cursor_type = SQL_CURSOR_FORWARD_ONLY;
 	opt->retrieve_data = SQL_RD_ON;
 	opt->use_bookmarks = SQL_UB_OFF;
-#if (ODBCVER >= 0x0300)
 	opt->metadata_id = SQL_FALSE;
-#endif /* ODBCVER */
-}
-
-
-/*
- * ARDFields initialize
- */
-void
-InitializeARDFields(ARDFields *opt)
-{
-	memset(opt, 0, sizeof(ARDFields));
-#if (ODBCVER >= 0x0300)
-	opt->size_of_rowset = 1;
-#endif /* ODBCVER */
-	opt->bind_size = 0;		/* default is to bind by column */
-	opt->size_of_rowset_odbc2 = 1;
-}
-/*
- * APDFields initialize
- */
-void
-InitializeAPDFields(APDFields *opt)
-{
-	memset(opt, 0, sizeof(APDFields));
-	opt->paramset_size = 1;
-	opt->param_bind_type = 0;	/* default is to bind by column */
 }
 
 
@@ -267,6 +241,7 @@ SC_Constructor(void)
 		rv->status = STMT_ALLOCATED;
 		rv->internal = FALSE;
 		rv->transition_status = 0;
+		rv->num_params = -1; /* unknown */
 
 		rv->__error_message = NULL;
 		rv->__error_number = 0;
@@ -303,10 +278,14 @@ SC_Constructor(void)
 
 		/* Clear Statement Options -- defaults will be set in AllocStmt */
 		memset(&rv->options, 0, sizeof(StatementOptions));
-		memset(&rv->ardopts, 0, sizeof(ARDFields));
-		memset(&rv->apdopts, 0, sizeof(APDFields));
-		memset(&rv->irdopts, 0, sizeof(IRDFields));
-		memset(&rv->ipdopts, 0, sizeof(IPDFields));
+		InitializeEmbeddedDescriptor((DescriptorClass *)&(rv->ardi),
+				rv, SQL_ATTR_APP_ROW_DESC);
+		InitializeEmbeddedDescriptor((DescriptorClass *)&(rv->apdi),
+				rv, SQL_ATTR_APP_PARAM_DESC);
+		InitializeEmbeddedDescriptor((DescriptorClass *)&(rv->irdi),
+				rv, SQL_ATTR_IMP_ROW_DESC);
+		InitializeEmbeddedDescriptor((DescriptorClass *)&(rv->ipdi),
+				rv, SQL_ATTR_IMP_PARAM_DESC);
 
 		rv->pre_executing = FALSE;
 		rv->inaccurate_result = FALSE;
@@ -315,60 +294,17 @@ SC_Constructor(void)
 		rv->error_recsize = -1;
 		rv->diag_row_count = 0;
 		rv->stmt_time = 0;
+		rv->execute_delegate = NULL;
+		rv->allocated_callbacks = 0;
+		rv->num_callbacks = 0;
+		rv->callbacks = NULL;
+		GetDataInfoInitialize(SC_get_GDTI(rv));
+		PutDataInfoInitialize(SC_get_PDTI(rv));
 		INIT_STMT_CS(rv);
 	}
 	return rv;
 }
 
-
-void ARDFields_free(ARDFields * self)
-{
-	if (self->bookmark)
-	{
-		free(self->bookmark);
-		self->bookmark = NULL;
-	}
-	/*
-	 * the memory pointed to by the bindings is not deallocated by the
-	 * driver but by the application that uses that driver, so we don't
-	 * have to care
-	 */
-	ARD_unbind_cols(self, TRUE);
-}
-
-void APDFields_free(APDFields * self)
-{
-	/* param bindings */
-	APD_free_params(self, STMT_FREE_PARAMS_ALL);
-}
-
-void IRDFields_free(IRDFields * self)
-{
-	/* Free the parsed field information */
-	if (self->fi)
-	{
-		int			i;
-
-		for (i = 0; i < (int) self->nfields; i++)
-		{
-			if (self->fi[i])
-			{
-				if (self->fi[i]->schema)
-					free(self->fi[i]->schema);
-				free(self->fi[i]);
-			}
-		}
-		free(self->fi);
-		self->fi = NULL;
-	}
-}
-
-void IPDFields_free(IPDFields * self)
-{
-	/* param bindings */
-	IPD_free_params(self, STMT_FREE_PARAMS_ALL);
-}
-
 char
 SC_Destructor(StatementClass *self)
 {
@@ -407,13 +343,19 @@ SC_Destructor(StatementClass *self)
 	}
 
 	/* Free the parsed field information */
-	ARDFields_free(&(self->ardopts));
-	APDFields_free(&(self->apdopts));
-	IRDFields_free(&(self->irdopts));
-	IPDFields_free(&(self->ipdopts));
+	DC_Destructor((DescriptorClass *) SC_get_ARDi(self));
+	DC_Destructor((DescriptorClass *) SC_get_APDi(self));
+	DC_Destructor((DescriptorClass *) SC_get_IRDi(self));
+	DC_Destructor((DescriptorClass *) SC_get_IPDi(self));
 	
 	if (self->__error_message)
 		free(self->__error_message);
+	cancelNeedDataState(self);
+	if (self->callbacks)
+		free(self->callbacks);
+    
+    GDATA_unbind_cols(SC_get_GDTI(self), TRUE);
+
 	DELETE_STMT_CS(self);
 	free(self);
 
@@ -430,8 +372,12 @@ SC_Destructor(StatementClass *self)
 void
 SC_free_params(StatementClass *self, char option)
 {
-	APD_free_params(SC_get_APD(self), option);
-	IPD_free_params(SC_get_IPD(self), option);
+	if (option != STMT_FREE_PARAMS_DATA_AT_EXEC_ONLY)
+	{
+		APD_free_params(SC_get_APDF(self), option);
+		IPD_free_params(SC_get_IPDF(self), option);
+	}
+	PDATA_free_params(SC_get_PDTI(self), option);
 	self->data_at_exec = -1;
 	self->current_exec_param = -1;
 	self->put_data = FALSE;
@@ -450,7 +396,7 @@ statement_type(const char *statement)
 	int			i;
 
 	/* ignore leading whitespace in query string */
-	while (*statement && (isspace((unsigned char) *statement) || *statement == '('))
+	while (*statement && (isspace((UCHAR) *statement) || *statement == '('))
 		statement++;
 
 	for (i = 0; Statement_Type[i].s; i++)
@@ -460,6 +406,18 @@ statement_type(const char *statement)
 	return STMT_TYPE_OTHER;
 }
 
+int
+SC_set_current_col(StatementClass *stmt, int col)
+{
+	if (col == stmt->current_col)
+		return col;
+	if (col >= 0)
+		reset_a_getdata_info(SC_get_GDTI(stmt), col + 1);
+	stmt->current_col = col;
+
+	return stmt->current_col;
+}
+
 void
 SC_set_prepared(StatementClass *stmt, BOOL prepared)
 {
@@ -468,15 +426,26 @@ SC_set_prepared(StatementClass *stmt, BOOL prepared)
 	if (!prepared)
 	{
 		ConnectionClass *conn = SC_get_conn(stmt);
-		if (CONN_CONNECTED == conn->status)
-		{
-			QResultClass	*res;
-			char dealloc_stmt[128];
 
-			sprintf(dealloc_stmt, "DEALLOCATE _PLAN%0x", stmt);
-			res = CC_send_query(conn, dealloc_stmt, NULL, 0);
-			if (res)
-				QR_Destructor(res); 
+		if (conn && CONN_CONNECTED == conn->status)
+		{
+			char	plannm[32];
+
+			sprintf(plannm, "_PLAN%0x", stmt);
+			if (CC_is_in_error_trans(conn))
+			{
+				CC_mark_a_plan_to_discard(conn, plannm);
+			}
+			else
+			{
+				QResultClass	*res;
+				char dealloc_stmt[128];
+
+				sprintf(dealloc_stmt, "DEALLOCATE \"%s\"", plannm);
+				res = CC_send_query(conn, dealloc_stmt, NULL, 0);
+				if (res)
+					QR_Destructor(res);
+			} 
 		}
 	}
 	stmt->prepared = prepared;
@@ -502,7 +471,11 @@ SC_initialize_stmts(StatementClass *self, BOOL initializeOriginal)
 			free(self->execute_statement);
 			self->execute_statement = NULL;
 		}
-		SC_set_prepared(self,FALSE);
+		self->prepare = FALSE;
+		SC_set_prepared(self, FALSE);
+		self->statement_type = STMT_TYPE_UNKNOWN; /* unknown */
+		self->status = STMT_READY;
+		self->num_params = -1; /* unknown */
 	}
 	if (self->stmt_with_params)
 	{
@@ -518,6 +491,47 @@ SC_initialize_stmts(StatementClass *self, BOOL initializeOriginal)
 	return 0;
 }
 
+BOOL	SC_opencheck(StatementClass *self, const char *func)
+{
+	QResultClass	*res;
+
+	if (self->status == STMT_EXECUTING)
+	{
+		SC_set_error(self, STMT_SEQUENCE_ERROR, "Statement is currently executing a transaction.");
+		return TRUE;
+	}
+	/*
+	 * We get here if a statement is prepared and executed to get the metadata.
+	 */
+	if (self->prepare && self->status == STMT_PREMATURE)
+	{
+		mylog("SC_opencheck: self->prepare && self->status == STMT_PREMATURE\n");
+		return FALSE;
+	}
+	
+	if (res = SC_get_Curres(self), NULL != res)
+	{
+		if (res->backend_tuples)
+		{
+			SC_set_error(self, STMT_SEQUENCE_ERROR, "The cursor is open.");
+			SC_log_error(func, "", self);
+			return TRUE;
+		}
+	}
+
+	return	FALSE;
+}
+
+RETCODE
+SC_initialize_and_recycle(StatementClass *self)
+{
+	SC_initialize_stmts(self, TRUE);
+	if (!SC_recycle_statement(self))
+		return SQL_ERROR;
+
+	return SQL_SUCCESS;
+}
+
 /*
  *	Called from SQLPrepare if STMT_PREMATURE, or
  *	from SQLExecute if STMT_FINISHED, or
@@ -583,7 +597,7 @@ SC_recycle_statement(StatementClass *self)
 		self->ntab = 0;
 	}
 	/* Free the parsed field information */
-	IRDFields_free(SC_get_IRD(self));
+	DC_Destructor((DescriptorClass *) SC_get_IRD(self));
 
 	self->parse_status = STMT_PARSE_NONE;
 	self->updatable = FALSE;
@@ -604,7 +618,7 @@ SC_recycle_statement(StatementClass *self)
 
 	self->currTuple = -1;
 	self->rowset_start = -1;
-	self->current_col = -1;
+	SC_set_current_col(self, -1);
 	self->bind_row = 0;
 	self->last_fetch_count = self->last_fetch_count_include_ommitted = 0;
 
@@ -621,6 +635,7 @@ SC_recycle_statement(StatementClass *self)
 	 */
 	SC_free_params(self, STMT_FREE_PARAMS_DATA_AT_EXEC_ONLY);
 	SC_initialize_stmts(self, FALSE);
+	cancelNeedDataState(self);
 	/*
 	 *	reset the current attr setting to the original one.
 	 */
@@ -648,15 +663,18 @@ SC_pre_execute(StatementClass *self)
 		if (self->statement_type == STMT_TYPE_SELECT)
 		{
 			char		old_pre_executing = self->pre_executing;
+			RETCODE		ret;
 
 			self->pre_executing = TRUE;
 			self->inaccurate_result = FALSE;
 
-			PGAPI_Execute(self);
+			ret = PGAPI_Execute(self, 0);
 
 			self->pre_executing = old_pre_executing;
 
-			if (self->status == STMT_FINISHED)
+			if (self->status == STMT_FINISHED &&
+			    (SQL_SUCCESS == ret ||
+			     SQL_SUCCESS_WITH_INFO == ret))
 			{
 				mylog("              preprocess: after status = FINISHED, so set PREMATURE\n");
 				self->status = STMT_PREMATURE;
@@ -677,11 +695,17 @@ SC_pre_execute(StatementClass *self)
 char
 SC_unbind_cols(StatementClass *self)
 {
-	ARDFields	*opts = SC_get_ARD(self);
+	ARDFields	*opts = SC_get_ARDF(self);
+	GetDataInfo	*gdata = SC_get_GDTI(self);
+	BindInfoClass	*bookmark;
 
 	ARD_unbind_cols(opts, FALSE);
-	opts->bookmark->buffer = NULL;
-	opts->bookmark->used = NULL;
+	GDATA_unbind_cols(gdata, FALSE);
+	if (bookmark = opts->bookmark, bookmark != NULL)
+	{
+		bookmark->buffer = NULL;
+		bookmark->used = NULL;
+	}
 
 	return 1;
 }
@@ -838,7 +862,7 @@ SC_get_time(StatementClass *stmt)
  *	just the current row number.  But it could be more sophisticated
  *	someday, such as mapping a key to a 32 bit value
  */
-unsigned long
+UInt4
 SC_get_bookmark(StatementClass *self)
 {
 	return (self->currTuple + 1);
@@ -851,6 +875,7 @@ SC_fetch(StatementClass *self)
 	CSTR func = "SC_fetch";
 	QResultClass *res = SC_get_Curres(self);
 	ARDFields	*opts;
+	GetDataInfo	*gdata;
 	int			retval,
 				result;
 
@@ -859,6 +884,7 @@ SC_fetch(StatementClass *self)
 	Oid			type;
 	char	   *value;
 	ColumnInfoClass *coli;
+	BindInfoClass	*bookmark;
 
 	/* TupleField *tupleField; */
 	ConnInfo   *ci = &(SC_get_conn(self)->connInfo);
@@ -921,32 +947,36 @@ SC_fetch(StatementClass *self)
 	self->last_fetch_count++;
 	self->last_fetch_count_include_ommitted++;
 
-	opts = SC_get_ARD(self);
+	opts = SC_get_ARDF(self);
 	/*
 	 * If the bookmark column was bound then return a bookmark. Since this
 	 * is used with SQLExtendedFetch, and the rowset size may be greater
 	 * than 1, and an application can use row or column wise binding, use
 	 * the code in copy_and_convert_field() to handle that.
 	 */
-	if (opts->bookmark->buffer)
+	if ((bookmark = opts->bookmark) && bookmark->buffer)
 	{
 		char		buf[32];
 		UInt4	offset = opts->row_offset_ptr ? *opts->row_offset_ptr : 0;
 
 		sprintf(buf, "%ld", SC_get_bookmark(self));
-		result = copy_and_convert_field(self, 0, buf,
-			 SQL_C_ULONG, opts->bookmark->buffer + offset, 0,
-			opts->bookmark->used ? opts->bookmark->used + (offset >> 2) : NULL);
+		SC_set_current_col(self, -1);
+		result = copy_and_convert_field(self, 0, buf, SQL_C_ULONG,
+						bookmark->buffer + offset, 0,
+						(SDWORD *) (bookmark->used ? bookmark->used + (offset >> 2) : NULL));
 	}
 
 	if (self->options.retrieve_data == SQL_RD_OFF)		/* data isn't required */
 		return SQL_SUCCESS;
+	gdata = SC_get_GDTI(self);
+	if (gdata->allocated != opts->allocated)
+		extend_getdata_info(gdata, opts->allocated, TRUE);
 	for (lf = 0; lf < num_cols; lf++)
 	{
 		mylog("fetch: cols=%d, lf=%d, opts = %u, opts->bindings = %u, buffer[] = %u\n", num_cols, lf, opts, opts->bindings, opts->bindings[lf].buffer);
 
 		/* reset for SQLGetData */
-		opts->bindings[lf].data_left = -1;
+		gdata->gdata[lf].data_left = -1;
 
 		if (opts->bindings[lf].buffer != NULL)
 		{
@@ -1023,19 +1053,15 @@ SC_fetch(StatementClass *self)
 	return result;
 }
 
+#define	return	DONT_CALL_RETURN_FROM_HERE???
 
 RETCODE
 SC_execute(StatementClass *self)
 {
 	CSTR func = "SC_execute";
 	ConnectionClass *conn;
-	APDFields	*apdopts;
-	char		was_ok, was_nonfatal, was_rows_affected = 1;
-	/* was_rows_affected is set to 0 iff an UPDATE or DELETE affects 
-	 * no rows. In this instance the driver should return
-	 * SQL_NO_DATA_FOUND instead of SQL_SUCCESS.
-	 */
- 
+	IPDFields	*ipdopts;
+	char		was_ok, was_nonfatal;
 	QResultClass	*res = NULL;
 	Int2		oldstatus,
 				numcols;
@@ -1043,6 +1069,7 @@ SC_execute(StatementClass *self)
 	ConnInfo   *ci;
 	UDWORD		qflag = 0;
 	BOOL		auto_begin = FALSE, is_in_trans;
+	int		func_cs_count = 0;
 
 
 	conn = SC_get_conn(self);
@@ -1058,14 +1085,12 @@ SC_execute(StatementClass *self)
 	 * 2) we are in autocommit off state and the statement isn't of type
 	 * OTHER.
 	 */
-	ENTER_CONN_CS(conn);
+	ENTER_INNER_CONN_CS(conn, func_cs_count);
 	if (CONN_EXECUTING == conn->status)
 	{
-		LEAVE_CONN_CS(conn);
 		SC_set_error(self, STMT_SEQUENCE_ERROR, "Connection is already in use.");
-		SC_log_error(func, "", self);
 		mylog("%s: problem with connection\n", func);
-		return SQL_ERROR;
+		goto cleanup;
 	}
 	is_in_trans = CC_is_in_trans(conn);
 	if (!self->internal && !is_in_trans &&
@@ -1078,10 +1103,8 @@ SC_execute(StatementClass *self)
 			qflag |= GO_INTO_TRANSACTION;
                 else if (!CC_begin(conn))
                 {
-			LEAVE_CONN_CS(conn);
 			SC_set_error(self, STMT_EXEC_ERROR, "Could not begin a transaction");
-                        SC_log_error(func, "", self);
-                        return SQL_ERROR;
+			goto cleanup;
                 }
 	}
 
@@ -1159,20 +1182,13 @@ SC_execute(StatementClass *self)
 	if (CONN_DOWN != conn->status)
 		conn->status = oldstatus;
 	self->status = STMT_FINISHED;
-	LEAVE_CONN_CS(conn);
+	LEAVE_INNER_CONN_CS(func_cs_count, conn);
 
 	/* Check the status of the result */
 	if (res)
 	{
 		was_ok = QR_command_successful(res);
 		was_nonfatal = QR_command_nonfatal(res);
-		if (res->command &&
-			(strncmp(res->command, "UPDATE", 6) == 0 ||
-			strncmp(res->command, "DELETE", 6) == 0) &&
-			strtoul(res->command + 7, NULL, 0) == 0)
-		{
-			was_rows_affected = 0;
-		}
 
 		if (was_ok)
 			SC_set_errornumber(self, STMT_OK);
@@ -1181,7 +1197,7 @@ SC_execute(StatementClass *self)
 
 		/* set cursor before the first tuple in the list */
 		self->currTuple = -1;
-		self->current_col = -1;
+		SC_set_current_col(self, -1);
 		self->rowset_start = -1;
 
 		/* issue "ABORT" when query aborted */
@@ -1202,14 +1218,13 @@ SC_execute(StatementClass *self)
 			/* now allocate the array to hold the binding info */
 			if (numcols > 0)
 			{
-				ARDFields	*opts = SC_get_ARD(self);
+				ARDFields	*opts = SC_get_ARDF(self);
 				extend_column_bindings(opts, numcols);
 				if (opts->bindings == NULL)
 				{
 					QR_Destructor(res);
 					SC_set_error(self, STMT_NO_MEMORY_ERROR,"Could not get enough free memory to store the binding information");
-					SC_log_error(func, "", self);
-					return SQL_ERROR;
+					goto cleanup;
 				}
 			}
 		}
@@ -1248,13 +1263,12 @@ SC_execute(StatementClass *self)
 		last->next = res;
 	}
 
-	apdopts = SC_get_APD(self);
+	ipdopts = SC_get_IPDF(self);
 	if (self->statement_type == STMT_TYPE_PROCCALL &&
 		(SC_get_errornumber(self) == STMT_OK ||
 		 SC_get_errornumber(self) == STMT_INFO_ONLY) &&
-		apdopts->parameters &&
-		apdopts->parameters[0].buffer &&
-		apdopts->parameters[0].paramType == SQL_PARAM_OUTPUT)
+		ipdopts->parameters &&
+		ipdopts->parameters[0].paramType == SQL_PARAM_OUTPUT)
 	{							/* get the return value of the procedure
 								 * call */
 		RETCODE		ret;
@@ -1263,7 +1277,13 @@ SC_execute(StatementClass *self)
 		ret = SC_fetch(hstmt);
 		if (ret == SQL_SUCCESS || ret == SQL_SUCCESS_WITH_INFO)
 		{
-			ret = PGAPI_GetData(hstmt, 1, apdopts->parameters[0].CType, apdopts->parameters[0].buffer, apdopts->parameters[0].buflen, apdopts->parameters[0].used);
+			APDFields	*apdopts = SC_get_APDF(self);
+
+			ret = PGAPI_GetData(hstmt, 1,
+					    apdopts->parameters[0].CType,
+					    apdopts->parameters[0].buffer,
+					    apdopts->parameters[0].buflen,
+					    (SDWORD *) apdopts->parameters[0].used);
 			if (ret != SQL_SUCCESS)
 			{
 				SC_set_error(self, STMT_EXEC_ERROR, "GetData to Procedure return failed.");
@@ -1274,11 +1294,11 @@ SC_execute(StatementClass *self)
 			SC_set_error(self, STMT_EXEC_ERROR, "SC_fetch to get a Procedure return failed.");
 		}
 	}
+#undef	return
+cleanup:
+	CLEANUP_FUNC_CONN_CS(func_cs_count, conn);
 	if (SC_get_errornumber(self) == STMT_OK)
-		if (was_rows_affected)
-			return SQL_SUCCESS;
-		else
-			return SQL_NO_DATA_FOUND;
+		return SQL_SUCCESS;
 	else if (SC_get_errornumber(self) == STMT_INFO_ONLY)
 		return SQL_SUCCESS_WITH_INFO;
 	else
@@ -1290,6 +1310,61 @@ SC_execute(StatementClass *self)
 	}
 }
 
+#define	CALLBACK_ALLOC_ONCE	4
+int enqueueNeedDataCallback(StatementClass *stmt, NeedDataCallfunc func, void *data)
+{
+	if (stmt->num_callbacks >= stmt->allocated_callbacks)
+	{
+		SC_REALLOC_return_with_error(stmt->callbacks, NeedDataCallback,
+			sizeof(NeedDataCallback) * (stmt->allocated_callbacks +
+				CALLBACK_ALLOC_ONCE), stmt,
+			 "Couldn't alloc callbacks", -1) 
+		stmt->allocated_callbacks += CALLBACK_ALLOC_ONCE;
+	}
+	stmt->callbacks[stmt->num_callbacks].func = func;
+	stmt->callbacks[stmt->num_callbacks].data = data;
+	stmt->num_callbacks++;
+
+inolog("enqueueNeedDataCallack stmt=%x, func=%x, count=%d\n", stmt, func, stmt->num_callbacks);
+	return stmt->num_callbacks;
+}
+
+RETCODE dequeueNeedDataCallback(RETCODE retcode, StatementClass *stmt)
+{
+	RETCODE			ret;
+	NeedDataCallfunc	func;
+	void			*data;
+	int			i, cnt;
+
+	mylog("dequeueNeedDataCallback ret=%d count=%d\n", retcode, stmt->num_callbacks);
+	if (SQL_NEED_DATA == retcode)
+		return retcode;
+	if (stmt->num_callbacks <= 0)
+		return retcode;
+	func = stmt->callbacks[0].func;
+	data = stmt->callbacks[0].data;
+	for (i = 1; i < stmt->num_callbacks; i++)
+		stmt->callbacks[i - 1] = stmt->callbacks[i];
+	cnt = --stmt->num_callbacks;
+	ret = (*func)(retcode, data);
+	free(data);
+	if (SQL_NEED_DATA != ret && cnt > 0)
+		ret = dequeueNeedDataCallback(ret, stmt);
+	return ret;
+}
+
+void	cancelNeedDataState(StatementClass *stmt)
+{
+	int	cnt = stmt->num_callbacks, i;
+
+	stmt->num_callbacks = 0;
+	for (i = 0; i < cnt; i++)
+	{
+		if (stmt->callbacks[i].data)
+			free(stmt->callbacks[i].data);
+	}
+	stmt->execute_delegate = NULL;
+}
 
 void
 SC_log_error(const char *func, const char *desc, const StatementClass *self)
@@ -1300,15 +1375,12 @@ SC_log_error(const char *func, const char *desc, const StatementClass *self)
 	if (self)
 	{
 		QResultClass *res = SC_get_Result(self);
-		const ARDFields	*opts = SC_get_ARD(self);
-		const APDFields	*apdopts = SC_get_APD(self);
+		const ARDFields	*opts = SC_get_ARDF(self);
+		const APDFields	*apdopts = SC_get_APDF(self);
 		int	rowsetSize;
 
-#if (ODBCVER >= 0x0300)
 		rowsetSize = (7 == self->transition_status ? opts->size_of_rowset_odbc2 : opts->size_of_rowset);
-#else
-		rowsetSize = opts->size_of_rowset_odbc2;
-#endif /* ODBCVER */
+
 		qlog("STATEMENT ERROR: func=%s, desc='%s', errnum=%d, errmsg='%s'\n", func, desc, self->__error_number, nullcheck(self->__error_message));
 		mylog("STATEMENT ERROR: func=%s, desc='%s', errnum=%d, errmsg='%s'\n", func, desc, self->__error_number, nullcheck(self->__error_message));
 		qlog("                 ------------------------------------------------------------\n");
